@@ -1,36 +1,10 @@
-# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-#
-#     http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
-# ==============================================================================
-
-"""DenseNet models for Keras.
-
-Reference:
-  - [Densely Connected Convolutional Networks](
-      https://arxiv.org/abs/1608.06993) (CVPR 2017)
-"""
-
-import tensorflow.compat.v2 as tf
-
 from keras.src import backend
+from keras.src import layers
+from keras.src.api_export import keras_export
 from keras.src.applications import imagenet_utils
-from keras.src.engine import training
-from keras.src.layers import VersionAwareLayers
-from keras.src.utils import data_utils
-from keras.src.utils import layer_utils
-
-# isort: off
-from tensorflow.python.util.tf_export import keras_export
+from keras.src.models import Functional
+from keras.src.ops import operation_utils
+from keras.src.utils import file_utils
 
 BASE_WEIGHTS_PATH = (
     "https://storage.googleapis.com/tensorflow/keras-applications/densenet/"
@@ -57,19 +31,17 @@ DENSENET201_WEIGHT_PATH_NO_TOP = (
     + "densenet201_weights_tf_dim_ordering_tf_kernels_notop.h5"
 )
 
-layers = VersionAwareLayers()
-
 
 def dense_block(x, blocks, name):
     """A dense block.
 
     Args:
-      x: input tensor.
-      blocks: integer, the number of building blocks.
-      name: string, block label.
+        x: input tensor.
+        blocks: integer, the number of building blocks.
+        name: string, block label.
 
     Returns:
-      Output tensor for the block.
+        Output tensor for the block.
     """
     for i in range(blocks):
         x = conv_block(x, 32, name=name + "_block" + str(i + 1))
@@ -80,12 +52,12 @@ def transition_block(x, reduction, name):
     """A transition block.
 
     Args:
-      x: input tensor.
-      reduction: float, compression rate at transition layers.
-      name: string, block label.
+        x: input tensor.
+        reduction: float, compression rate at transition layers.
+        name: string, block label.
 
     Returns:
-      output tensor for the block.
+        Output tensor for the block.
     """
     bn_axis = 3 if backend.image_data_format() == "channels_last" else 1
     x = layers.BatchNormalization(
@@ -93,7 +65,7 @@ def transition_block(x, reduction, name):
     )(x)
     x = layers.Activation("relu", name=name + "_relu")(x)
     x = layers.Conv2D(
-        int(backend.int_shape(x)[bn_axis] * reduction),
+        int(x.shape[bn_axis] * reduction),
         1,
         use_bias=False,
         name=name + "_conv",
@@ -106,12 +78,12 @@ def conv_block(x, growth_rate, name):
     """A building block for a dense block.
 
     Args:
-      x: input tensor.
-      growth_rate: float, growth rate at dense layers.
-      name: string, block label.
+        x: input tensor.
+        growth_rate: float, growth rate at dense layers.
+        name: string, block label.
 
     Returns:
-      Output tensor for the block.
+        Output tensor for the block.
     """
     bn_axis = 3 if backend.image_data_format() == "channels_last" else 1
     x1 = layers.BatchNormalization(
@@ -141,6 +113,7 @@ def DenseNet(
     pooling=None,
     classes=1000,
     classifier_activation="softmax",
+    name="densenet",
 ):
     """Instantiates the DenseNet architecture.
 
@@ -160,52 +133,62 @@ def DenseNet(
       https://keras.io/guides/transfer_learning/).
 
     Note: each Keras Application expects a specific kind of input preprocessing.
-    For DenseNet, call `tf.keras.applications.densenet.preprocess_input` on your
-    inputs before passing them to the model.
+    For DenseNet, call `keras.applications.densenet.preprocess_input`
+    on your inputs before passing them to the model.
     `densenet.preprocess_input` will scale pixels between 0 and 1 and then
-    will normalize each channel with respect to the ImageNet dataset statistics.
+    will normalize each channel with respect to the ImageNet
+    dataset statistics.
 
     Args:
-      blocks: numbers of building blocks for the four dense layers.
-      include_top: whether to include the fully-connected
-        layer at the top of the network.
-      weights: one of `None` (random initialization),
-        'imagenet' (pre-training on ImageNet),
-        or the path to the weights file to be loaded.
-      input_tensor: optional Keras tensor
-        (i.e. output of `layers.Input()`)
-        to use as image input for the model.
-      input_shape: optional shape tuple, only to be specified
-        if `include_top` is False (otherwise the input shape
-        has to be `(224, 224, 3)` (with `'channels_last'` data format)
-        or `(3, 224, 224)` (with `'channels_first'` data format).
-        It should have exactly 3 inputs channels,
-        and width and height should be no smaller than 32.
-        E.g. `(200, 200, 3)` would be one valid value.
-      pooling: optional pooling mode for feature extraction
-        when `include_top` is `False`.
-        - `None` means that the output of the model will be
-            the 4D tensor output of the
-            last convolutional block.
-        - `avg` means that global average pooling
-            will be applied to the output of the
-            last convolutional block, and thus
-            the output of the model will be a 2D tensor.
-        - `max` means that global max pooling will
-            be applied.
-      classes: optional number of classes to classify images
-        into, only to be specified if `include_top` is True, and
-        if no `weights` argument is specified.
-      classifier_activation: A `str` or callable. The activation function to use
-        on the "top" layer. Ignored unless `include_top=True`. Set
-        `classifier_activation=None` to return the logits of the "top" layer.
-        When loading pretrained weights, `classifier_activation` can only
-        be `None` or `"softmax"`.
+        blocks: numbers of building blocks for the four dense layers.
+        include_top: whether to include the fully-connected
+            layer at the top of the network.
+        weights: one of `None` (random initialization),
+            `"imagenet"` (pre-training on ImageNet),
+            or the path to the weights file to be loaded.
+        input_tensor: optional Keras tensor
+            (i.e. output of `layers.Input()`)
+            to use as image input for the model.
+        input_shape: optional shape tuple, only to be specified
+            if `include_top` is False (otherwise the input shape
+            has to be `(224, 224, 3)`
+            (with `'channels_last'` data format)
+            or `(3, 224, 224)` (with `'channels_first'` data format).
+            It should have exactly 3 inputs channels,
+            and width and height should be no smaller than 32.
+            E.g. `(200, 200, 3)` would be one valid value.
+        pooling: optional pooling mode for feature extraction
+            when `include_top` is `False`.
+            - `None` means that the output of the model will be
+                the 4D tensor output of the
+                last convolutional block.
+            - `avg` means that global average pooling
+                will be applied to the output of the
+                last convolutional block, and thus
+                the output of the model will be a 2D tensor.
+            - `max` means that global max pooling will
+                be applied.
+        classes: optional number of classes to classify images
+            into, only to be specified if `include_top` is `True`, and
+            if no `weights` argument is specified.
+        classifier_activation: A `str` or callable.
+            The activation function to use
+            on the "top" layer. Ignored unless `include_top=True`. Set
+            `classifier_activation=None` to return the logits of the "top"
+            layer. When loading pretrained weights, `classifier_activation`
+            can only be `None` or `"softmax"`.
+        name: The name of the model (string).
 
     Returns:
-      A `keras.Model` instance.
+        A model instance.
     """
-    if not (weights in {"imagenet", None} or tf.io.gfile.exists(weights)):
+    if backend.image_data_format() == "channels_first":
+        raise ValueError(
+            "DenseNet does not support the `channels_first` image data "
+            "format. Switch to `channels_last` by editing your local "
+            "config file at ~/.keras/keras.json"
+        )
+    if not (weights in {"imagenet", None} or file_utils.exists(weights)):
         raise ValueError(
             "The `weights` argument should be either "
             "`None` (random initialization), `imagenet` "
@@ -240,11 +223,11 @@ def DenseNet(
     bn_axis = 3 if backend.image_data_format() == "channels_last" else 1
 
     x = layers.ZeroPadding2D(padding=((3, 3), (3, 3)))(img_input)
-    x = layers.Conv2D(64, 7, strides=2, use_bias=False, name="conv1/conv")(x)
+    x = layers.Conv2D(64, 7, strides=2, use_bias=False, name="conv1_conv")(x)
     x = layers.BatchNormalization(
-        axis=bn_axis, epsilon=1.001e-5, name="conv1/bn"
+        axis=bn_axis, epsilon=1.001e-5, name="conv1_bn"
     )(x)
-    x = layers.Activation("relu", name="conv1/relu")(x)
+    x = layers.Activation("relu", name="conv1_relu")(x)
     x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)))(x)
     x = layers.MaxPooling2D(3, strides=2, name="pool1")(x)
 
@@ -275,39 +258,32 @@ def DenseNet(
     # Ensure that the model takes into account
     # any potential predecessors of `input_tensor`.
     if input_tensor is not None:
-        inputs = layer_utils.get_source_inputs(input_tensor)
+        inputs = operation_utils.get_source_inputs(input_tensor)
     else:
         inputs = img_input
 
     # Create model.
-    if blocks == [6, 12, 24, 16]:
-        model = training.Model(inputs, x, name="densenet121")
-    elif blocks == [6, 12, 32, 32]:
-        model = training.Model(inputs, x, name="densenet169")
-    elif blocks == [6, 12, 48, 32]:
-        model = training.Model(inputs, x, name="densenet201")
-    else:
-        model = training.Model(inputs, x, name="densenet")
+    model = Functional(inputs, x, name=name)
 
     # Load weights.
     if weights == "imagenet":
         if include_top:
             if blocks == [6, 12, 24, 16]:
-                weights_path = data_utils.get_file(
+                weights_path = file_utils.get_file(
                     "densenet121_weights_tf_dim_ordering_tf_kernels.h5",
                     DENSENET121_WEIGHT_PATH,
                     cache_subdir="models",
                     file_hash="9d60b8095a5708f2dcce2bca79d332c7",
                 )
             elif blocks == [6, 12, 32, 32]:
-                weights_path = data_utils.get_file(
+                weights_path = file_utils.get_file(
                     "densenet169_weights_tf_dim_ordering_tf_kernels.h5",
                     DENSENET169_WEIGHT_PATH,
                     cache_subdir="models",
                     file_hash="d699b8f76981ab1b30698df4c175e90b",
                 )
             elif blocks == [6, 12, 48, 32]:
-                weights_path = data_utils.get_file(
+                weights_path = file_utils.get_file(
                     "densenet201_weights_tf_dim_ordering_tf_kernels.h5",
                     DENSENET201_WEIGHT_PATH,
                     cache_subdir="models",
@@ -315,21 +291,21 @@ def DenseNet(
                 )
         else:
             if blocks == [6, 12, 24, 16]:
-                weights_path = data_utils.get_file(
+                weights_path = file_utils.get_file(
                     "densenet121_weights_tf_dim_ordering_tf_kernels_notop.h5",
                     DENSENET121_WEIGHT_PATH_NO_TOP,
                     cache_subdir="models",
                     file_hash="30ee3e1110167f948a6b9946edeeb738",
                 )
             elif blocks == [6, 12, 32, 32]:
-                weights_path = data_utils.get_file(
+                weights_path = file_utils.get_file(
                     "densenet169_weights_tf_dim_ordering_tf_kernels_notop.h5",
                     DENSENET169_WEIGHT_PATH_NO_TOP,
                     cache_subdir="models",
                     file_hash="b8c4d4c20dd625c148057b9ff1c1176b",
                 )
             elif blocks == [6, 12, 48, 32]:
-                weights_path = data_utils.get_file(
+                weights_path = file_utils.get_file(
                     "densenet201_weights_tf_dim_ordering_tf_kernels_notop.h5",
                     DENSENET201_WEIGHT_PATH_NO_TOP,
                     cache_subdir="models",
@@ -343,7 +319,10 @@ def DenseNet(
 
 
 @keras_export(
-    "keras.applications.densenet.DenseNet121", "keras.applications.DenseNet121"
+    [
+        "keras.applications.densenet.DenseNet121",
+        "keras.applications.DenseNet121",
+    ]
 )
 def DenseNet121(
     include_top=True,
@@ -353,6 +332,7 @@ def DenseNet121(
     pooling=None,
     classes=1000,
     classifier_activation="softmax",
+    name="densenet121",
 ):
     """Instantiates the Densenet121 architecture."""
     return DenseNet(
@@ -364,11 +344,15 @@ def DenseNet121(
         pooling,
         classes,
         classifier_activation,
+        name=name,
     )
 
 
 @keras_export(
-    "keras.applications.densenet.DenseNet169", "keras.applications.DenseNet169"
+    [
+        "keras.applications.densenet.DenseNet169",
+        "keras.applications.DenseNet169",
+    ]
 )
 def DenseNet169(
     include_top=True,
@@ -378,6 +362,7 @@ def DenseNet169(
     pooling=None,
     classes=1000,
     classifier_activation="softmax",
+    name="densenet169",
 ):
     """Instantiates the Densenet169 architecture."""
     return DenseNet(
@@ -389,11 +374,15 @@ def DenseNet169(
         pooling,
         classes,
         classifier_activation,
+        name=name,
     )
 
 
 @keras_export(
-    "keras.applications.densenet.DenseNet201", "keras.applications.DenseNet201"
+    [
+        "keras.applications.densenet.DenseNet201",
+        "keras.applications.DenseNet201",
+    ]
 )
 def DenseNet201(
     include_top=True,
@@ -403,6 +392,7 @@ def DenseNet201(
     pooling=None,
     classes=1000,
     classifier_activation="softmax",
+    name="densenet201",
 ):
     """Instantiates the Densenet201 architecture."""
     return DenseNet(
@@ -414,6 +404,7 @@ def DenseNet201(
         pooling,
         classes,
         classifier_activation,
+        name=name,
     )
 
 
@@ -438,58 +429,60 @@ decode_predictions.__doc__ = imagenet_utils.decode_predictions.__doc__
 
 DOC = """
 
-  Reference:
-  - [Densely Connected Convolutional Networks](
-      https://arxiv.org/abs/1608.06993) (CVPR 2017)
+Reference:
+- [Densely Connected Convolutional Networks](
+    https://arxiv.org/abs/1608.06993) (CVPR 2017)
 
-  Optionally loads weights pre-trained on ImageNet.
-  Note that the data format convention used by the model is
-  the one specified in your Keras config at `~/.keras/keras.json`.
+Optionally loads weights pre-trained on ImageNet.
+Note that the data format convention used by the model is
+the one specified in your Keras config at `~/.keras/keras.json`.
 
-  Note: each Keras Application expects a specific kind of input preprocessing.
-  For DenseNet, call `tf.keras.applications.densenet.preprocess_input` on your
-  inputs before passing them to the model.
+Note: each Keras Application expects a specific kind of input preprocessing.
+For DenseNet, call `keras.applications.densenet.preprocess_input`
+on your inputs before passing them to the model.
 
-  Args:
+Args:
     include_top: whether to include the fully-connected
-      layer at the top of the network.
+        layer at the top of the network.
     weights: one of `None` (random initialization),
-      'imagenet' (pre-training on ImageNet),
-      or the path to the weights file to be loaded.
-    input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
-      to use as image input for the model.
+        `"imagenet"` (pre-training on ImageNet),
+        or the path to the weights file to be loaded.
+        input_tensor: optional Keras tensor
+        (i.e. output of `layers.Input()`)
+        to use as image input for the model.
     input_shape: optional shape tuple, only to be specified
-      if `include_top` is False (otherwise the input shape
-      has to be `(224, 224, 3)` (with `'channels_last'` data format)
-      or `(3, 224, 224)` (with `'channels_first'` data format).
-      It should have exactly 3 inputs channels,
-      and width and height should be no smaller than 32.
-      E.g. `(200, 200, 3)` would be one valid value.
+        if `include_top` is False (otherwise the input shape
+        has to be `(224, 224, 3)` (with `'channels_last'` data format)
+        or `(3, 224, 224)` (with `'channels_first'` data format).
+        It should have exactly 3 inputs channels,
+        and width and height should be no smaller than 32.
+        E.g. `(200, 200, 3)` would be one valid value.
     pooling: Optional pooling mode for feature extraction
-      when `include_top` is `False`.
-      - `None` means that the output of the model will be
-          the 4D tensor output of the
-          last convolutional block.
-      - `avg` means that global average pooling
-          will be applied to the output of the
-          last convolutional block, and thus
-          the output of the model will be a 2D tensor.
-      - `max` means that global max pooling will
-          be applied.
+        when `include_top` is `False`.
+        - `None` means that the output of the model will be
+            the 4D tensor output of the
+            last convolutional block.
+        - `avg` means that global average pooling
+            will be applied to the output of the
+            last convolutional block, and thus
+            the output of the model will be a 2D tensor.
+        - `max` means that global max pooling will
+            be applied.
     classes: optional number of classes to classify images
-      into, only to be specified if `include_top` is True, and
-      if no `weights` argument is specified.
-    classifier_activation: A `str` or callable. The activation function to use
-      on the "top" layer. Ignored unless `include_top=True`. Set
-      `classifier_activation=None` to return the logits of the "top" layer.
-      When loading pretrained weights, `classifier_activation` can only
-      be `None` or `"softmax"`.
+        into, only to be specified if `include_top` is `True`, and
+        if no `weights` argument is specified.
+    classifier_activation: A `str` or callable.
+        The activation function to use
+        on the "top" layer. Ignored unless `include_top=True`. Set
+        `classifier_activation=None` to return the logits
+        of the "top" layer. When loading pretrained weights,
+        `classifier_activation` can only be `None` or `"softmax"`.
+    name: The name of the model (string).
 
-  Returns:
+Returns:
     A Keras model instance.
 """
 
 setattr(DenseNet121, "__doc__", DenseNet121.__doc__ + DOC)
 setattr(DenseNet169, "__doc__", DenseNet169.__doc__ + DOC)
 setattr(DenseNet201, "__doc__", DenseNet201.__doc__ + DOC)
-
