@@ -1,30 +1,24 @@
-# Copyright 2023 The TensorFlow Authors. All Rights Reserved.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-#
-#     http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
-# ==============================================================================
 """Library for exporting inference-only Keras models/layers."""
 
-import tensorflow.compat.v2 as tf
-from tensorflow.python.util.tf_export import keras_export
+import inspect
+import itertools
+import string
 
-from keras.src.engine import base_layer
-from keras.src.engine import functional
-from keras.src.engine import sequential
+from absl import logging
+
+from keras.src import backend
+from keras.src import tree
+from keras.src.api_export import keras_export
+from keras.src.backend.common.stateless_scope import StatelessScope
+from keras.src.layers import Layer
+from keras.src.models import Functional
+from keras.src.models import Sequential
 from keras.src.utils import io_utils
+from keras.src.utils.module_utils import tensorflow as tf
 
 
 @keras_export("keras.export.ExportArchive")
-class ExportArchive(tf.__internal__.tracking.AutoTrackable):
+class ExportArchive:
     """ExportArchive is used to write SavedModel artifacts (e.g. for inference).
 
     If you have a Keras model or layer that you want to export as SavedModel for
@@ -94,21 +88,67 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
         self._endpoint_names = []
         self._endpoint_signatures = {}
         self.tensorflow_version = tf.__version__
-        self.variables = []
-        self.trainable_variables = []
-        self.non_trainable_variables = []
 
-    @tf.__internal__.tracking.no_automatic_dependency_tracking
+        self._tf_trackable = tf.__internal__.tracking.AutoTrackable()
+        self._tf_trackable.variables = []
+        self._tf_trackable.trainable_variables = []
+        self._tf_trackable.non_trainable_variables = []
+
+        if backend.backend() == "jax":
+            self._backend_variables = []
+            self._backend_trainable_variables = []
+            self._backend_non_trainable_variables = []
+
+        if backend.backend() not in ("tensorflow", "jax"):
+            raise NotImplementedError(
+                "The export API is only compatible with JAX and TF backends."
+            )
+
+    @property
+    def variables(self):
+        return self._tf_trackable.variables
+
+    @property
+    def trainable_variables(self):
+        return self._tf_trackable.trainable_variables
+
+    @property
+    def non_trainable_variables(self):
+        return self._tf_trackable.non_trainable_variables
+
     def track(self, resource):
-        """Track the variables (and other assets) of a layer or model."""
-        if not isinstance(resource, tf.__internal__.tracking.Trackable):
+        """Track the variables (and other assets) of a layer or model.
+
+        By default, all variables used by an endpoint function
+        are automatically tracked when you call `add_endpoint()`.
+        However, non-variables assets such as lookup tables
+        need to be tracked manually. Note that lookup tables
+        used by built-in Keras layers
+        (`TextVectorization`, `IntegerLookup`, `StringLookup`)
+        are automatically tracked in `add_endpoint()`.
+
+        Arguments:
+            resource: A trackable TensorFlow resource.
+        """
+        if backend.backend() == "tensorflow" and not isinstance(
+            resource, tf.__internal__.tracking.Trackable
+        ):
             raise ValueError(
                 "Invalid resource type. Expected an instance of a "
                 "TensorFlow `Trackable` (such as a Keras `Layer` or `Model`). "
                 f"Received instead an object of type '{type(resource)}'. "
                 f"Object received: {resource}"
             )
-        if isinstance(resource, base_layer.Layer):
+        if backend.backend() == "jax" and not isinstance(
+            resource, backend.jax.layer.JaxLayer
+        ):
+            raise ValueError(
+                "Invalid resource type. Expected an instance of a "
+                "JAX-based Keras `Layer` or `Model`. "
+                f"Received instead an object of type '{type(resource)}'. "
+                f"Object received: {resource}"
+            )
+        if isinstance(resource, Layer):
             if not resource.built:
                 raise ValueError(
                     "The layer provided has not yet been built. "
@@ -122,14 +162,42 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
             self._tracked = []
         self._tracked.append(resource)
 
-        if isinstance(resource, base_layer.Layer):
+        if isinstance(resource, Layer):
             # Variables in the lists below are actually part of the trackables
             # that get saved, because the lists are created in __init__.
-            self.variables += resource.variables
-            self.trainable_variables += resource.trainable_variables
-            self.non_trainable_variables += resource.non_trainable_variables
+            if backend.backend() == "jax":
 
-    def add_endpoint(self, name, fn, input_signature=None):
+                trainable_variables = tree.flatten(resource.trainable_variables)
+                non_trainable_variables = tree.flatten(
+                    resource.non_trainable_variables
+                )
+                self._backend_trainable_variables += trainable_variables
+                self._backend_non_trainable_variables += non_trainable_variables
+                self._backend_variables = (
+                    self._backend_trainable_variables
+                    + self._backend_non_trainable_variables
+                )
+
+                self._tf_trackable.trainable_variables += [
+                    tf.Variable(v) for v in trainable_variables
+                ]
+                self._tf_trackable.non_trainable_variables += [
+                    tf.Variable(v) for v in non_trainable_variables
+                ]
+                self._tf_trackable.variables = (
+                    self._tf_trackable.trainable_variables
+                    + self._tf_trackable.non_trainable_variables
+                )
+            else:
+                self._tf_trackable.variables += resource.variables
+                self._tf_trackable.trainable_variables += (
+                    resource.trainable_variables
+                )
+                self._tf_trackable.non_trainable_variables += (
+                    resource.non_trainable_variables
+                )
+
+    def add_endpoint(self, name, fn, input_signature=None, jax2tf_kwargs=None):
         """Register a new serving endpoint.
 
         Arguments:
@@ -149,6 +217,15 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
                 per positional input argument of `fn`). Nested arguments are
                 allowed (see below for an example showing a Functional model
                 with 2 input arguments).
+            jax2tf_kwargs: Optional. A dict for arguments to pass to `jax2tf`.
+                Supported only when the backend is JAX. See documentation for
+                [`jax2tf.convert`](
+                    https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md).
+                The values for `native_serialization` and `polymorphic_shapes`,
+                if not provided, are automatically computed.
+
+        Returns:
+            The `tf.function` wrapping `fn` that was added to the archive.
 
         Example:
 
@@ -239,8 +316,72 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
         if name in self._endpoint_names:
             raise ValueError(f"Endpoint name '{name}' is already taken.")
 
+        if jax2tf_kwargs and backend.backend() != "jax":
+            raise ValueError(
+                "'jax2tf_kwargs' is only supported with the jax backend. "
+                f"Current backend: {backend.backend()}"
+            )
+
         if input_signature:
-            decorated_fn = tf.function(fn, input_signature=input_signature)
+            if backend.backend() == "tensorflow":
+                decorated_fn = tf.function(
+                    fn, input_signature=input_signature, autograph=False
+                )
+            else:  # JAX backend
+
+                # 1. Create a stateless wrapper for `fn`
+                # 2. jax2tf the stateless wrapper
+                # 3. Create a stateful function that binds the variables with
+                #    the jax2tf converted stateless wrapper
+                # 4. Make the signature of the stateful function the same as the
+                #    original function
+                # 5. Wrap in a `tf.function`
+                def stateless_fn(variables, *args, **kwargs):
+                    state_mapping = zip(self._backend_variables, variables)
+                    with StatelessScope(state_mapping=state_mapping) as scope:
+                        output = fn(*args, **kwargs)
+
+                    # Gather updated non-trainable variables
+                    non_trainable_variables = []
+                    for var in self._backend_non_trainable_variables:
+                        new_value = scope.get_current_value(var)
+                        non_trainable_variables.append(new_value)
+                    return output, non_trainable_variables
+
+                jax2tf_stateless_fn = self._convert_jax2tf_function(
+                    stateless_fn,
+                    input_signature,
+                    jax2tf_kwargs=jax2tf_kwargs,
+                )
+
+                def stateful_fn(*args, **kwargs):
+                    output, non_trainable_variables = jax2tf_stateless_fn(
+                        # Change the trackable `ListWrapper` to a plain `list`
+                        list(self._tf_trackable.variables),
+                        *args,
+                        **kwargs,
+                    )
+                    for var, new_value in zip(
+                        self._tf_trackable.non_trainable_variables,
+                        non_trainable_variables,
+                    ):
+                        var.assign(new_value)
+                    return output
+
+                # Note: we truncate the number of parameters to what is
+                # specified by `input_signature`.
+                fn_signature = inspect.signature(fn)
+                fn_parameters = list(fn_signature.parameters.values())
+                stateful_fn.__signature__ = inspect.Signature(
+                    parameters=fn_parameters[0 : len(input_signature)],
+                    return_annotation=fn_signature.return_annotation,
+                )
+
+                decorated_fn = tf.function(
+                    stateful_fn,
+                    input_signature=input_signature,
+                    autograph=False,
+                )
             self._endpoint_signatures[name] = input_signature
         else:
             if isinstance(fn, tf.types.experimental.GenericFunction):
@@ -272,8 +413,9 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
                     "    ],\n"
                     ")"
                 )
-        setattr(self, name, decorated_fn)
+        setattr(self._tf_trackable, name, decorated_fn)
         self._endpoint_names.append(name)
+        return decorated_fn
 
     def add_variable_collection(self, name, variables):
         """Register a set of variables to be retrieved after reloading.
@@ -309,13 +451,19 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
                 "Expected `variables` to be a list/tuple/set. "
                 f"Received instead object of type '{type(variables)}'."
             )
-        if not all(isinstance(v, tf.Variable) for v in variables):
+        # Ensure that all variables added are either tf.Variables
+        # or Variables created by Keras 3 with the TF or JAX backends.
+        if not all(
+            isinstance(v, (tf.Variable, backend.Variable)) for v in variables
+        ):
             raise ValueError(
                 "Expected all elements in `variables` to be "
                 "`tf.Variable` instances. Found instead the following types: "
                 f"{list(set(type(v) for v in variables))}"
             )
-        setattr(self, name, list(variables))
+        if backend.backend() == "jax":
+            variables = tree.flatten(tree.map_structure(tf.Variable, variables))
+        setattr(self._tf_trackable, name, list(variables))
 
     def write_out(self, filepath, options=None):
         """Write the corresponding SavedModel to disk.
@@ -337,7 +485,8 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
             raise ValueError(
                 "No endpoints have been set yet. Call add_endpoint()."
             )
-        self._filter_and_track_resources()
+        if backend.backend() == "tensorflow":
+            self._filter_and_track_resources()
 
         signatures = {}
         for name in self._endpoint_names:
@@ -347,12 +496,17 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
             signatures["serving_default"] = self._get_concrete_fn(
                 self._endpoint_names[0]
             )
+
         tf.saved_model.save(
-            self, filepath, options=options, signatures=signatures
+            self._tf_trackable,
+            filepath,
+            options=options,
+            signatures=signatures,
         )
+
         # Print out available endpoints
         endpoints = "\n\n".join(
-            _print_signature(getattr(self, name), name)
+            _print_signature(getattr(self._tf_trackable, name), name)
             for name in self._endpoint_names
         )
         io_utils.print_msg(
@@ -364,9 +518,11 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
     def _get_concrete_fn(self, endpoint):
         """Workaround for some SavedModel quirks."""
         if endpoint in self._endpoint_signatures:
-            return getattr(self, endpoint)
+            return getattr(self._tf_trackable, endpoint)
         else:
-            traces = getattr(self, endpoint)._trackable_children("saved_model")
+            traces = getattr(self._tf_trackable, endpoint)._trackable_children(
+                "saved_model"
+            )
             return list(traces.values())[0]
 
     def _get_variables_used_by_endpoints(self):
@@ -378,42 +534,148 @@ class ExportArchive(tf.__internal__.tracking.AutoTrackable):
         # Start by extracting variables from endpoints.
         fns = [self._get_concrete_fn(name) for name in self._endpoint_names]
         tvs, ntvs = _list_variables_used_by_fns(fns)
-        self._all_variables = list(tvs + ntvs)
+        self._tf_trackable._all_variables = list(tvs + ntvs)
 
         # Next, track lookup tables.
         # Hopefully, one day this will be automated at the tf.function level.
-        self._misc_assets = []
-        from keras.src.layers.preprocessing.index_lookup import IndexLookup
+        self._tf_trackable._misc_assets = []
+        from keras.src.layers import IntegerLookup
+        from keras.src.layers import StringLookup
+        from keras.src.layers import TextVectorization
 
         if hasattr(self, "_tracked"):
             for root in self._tracked:
                 descendants = tf.train.TrackableView(root).descendants()
                 for trackable in descendants:
-                    if isinstance(trackable, IndexLookup):
-                        self._misc_assets.append(trackable)
+                    if isinstance(
+                        trackable,
+                        (IntegerLookup, StringLookup, TextVectorization),
+                    ):
+                        self._tf_trackable._misc_assets.append(trackable)
+
+    def _convert_jax2tf_function(self, fn, input_signature, jax2tf_kwargs=None):
+        from jax.experimental import jax2tf
+
+        if jax2tf_kwargs is None:
+            jax2tf_kwargs = {}
+
+        if "native_serialization" not in jax2tf_kwargs:
+            jax2tf_kwargs["native_serialization"] = (
+                self._check_device_compatible()
+            )
+
+        variables_shapes = self._to_polymorphic_shape(
+            self._backend_variables, allow_none=False
+        )
+        if "polymorphic_shapes" in jax2tf_kwargs:
+            input_shapes = jax2tf_kwargs["polymorphic_shapes"]
+        else:
+            input_shapes = self._to_polymorphic_shape(input_signature)
+        jax2tf_kwargs["polymorphic_shapes"] = [variables_shapes] + input_shapes
+
+        return jax2tf.convert(fn, **jax2tf_kwargs)
+
+    def _to_polymorphic_shape(self, struct, allow_none=True):
+        if allow_none:
+            # Generates unique names: a, b, ... z, aa, ab, ... az, ba, ... zz
+            # for unknown non-batch dims. Defined here to be scope per endpoint.
+            dim_names = itertools.chain(
+                string.ascii_lowercase,
+                itertools.starmap(
+                    lambda a, b: a + b,
+                    itertools.product(string.ascii_lowercase, repeat=2),
+                ),
+            )
+
+        def convert_shape(x):
+            poly_shape = []
+            for index, dim in enumerate(list(x.shape)):
+                if dim is not None:
+                    poly_shape.append(str(dim))
+                elif not allow_none:
+                    raise ValueError(
+                        f"Illegal None dimension in {x} with shape {x.shape}"
+                    )
+                elif index == 0:
+                    poly_shape.append("batch")
+                else:
+                    poly_shape.append(next(dim_names))
+            return "(" + ", ".join(poly_shape) + ")"
+
+        return tree.map_structure(convert_shape, struct)
+
+    def _check_device_compatible(self):
+        from jax import default_backend as jax_device
+
+        if (
+            jax_device() == "gpu"
+            and len(tf.config.list_physical_devices("GPU")) == 0
+        ):
+            logging.warning(
+                "JAX backend is using GPU for export, but installed "
+                "TF package cannot access GPU, so reloading the model with "
+                "the TF runtime in the same environment will not work. "
+                "To use JAX-native serialization for high-performance export "
+                "and serving, please install `tensorflow-gpu` and ensure "
+                "CUDA version compatiblity between your JAX and TF "
+                "installations."
+            )
+            return False
+        else:
+            return True
 
 
 def export_model(model, filepath):
     export_archive = ExportArchive()
     export_archive.track(model)
-    if isinstance(model, (functional.Functional, sequential.Sequential)):
-        input_signature = tf.nest.map_structure(_make_tensor_spec, model.inputs)
+    if isinstance(model, (Functional, Sequential)):
+        input_signature = tree.map_structure(_make_tensor_spec, model.inputs)
         if isinstance(input_signature, list) and len(input_signature) > 1:
             input_signature = [input_signature]
         export_archive.add_endpoint("serve", model.__call__, input_signature)
     else:
-        save_spec = model._get_save_spec()
-        if not save_spec:
+        input_signature = _get_input_signature(model)
+        if not input_signature or not model._called:
             raise ValueError(
                 "The model provided has never called. "
                 "It must be called at least once before export."
             )
-        input_signature = [save_spec]
         export_archive.add_endpoint("serve", model.__call__, input_signature)
     export_archive.write_out(filepath)
 
 
-class ReloadedLayer(base_layer.Layer):
+def _get_input_signature(model):
+    shapes_dict = getattr(model, "_build_shapes_dict", None)
+    if not shapes_dict:
+        return None
+
+    def make_tensor_spec(structure):
+        # We need to turn wrapper structures like TrackingDict or _DictWrapper
+        # into plain Python structures because they don't work with jax2tf/JAX.
+        if isinstance(structure, dict):
+            return {k: make_tensor_spec(v) for k, v in structure.items()}
+        elif isinstance(structure, tuple):
+            if all(isinstance(d, (int, type(None))) for d in structure):
+                return tf.TensorSpec(
+                    shape=(None,) + structure[1:], dtype=model.input_dtype
+                )
+            return tuple(make_tensor_spec(v) for v in structure)
+        elif isinstance(structure, list):
+            if all(isinstance(d, (int, type(None))) for d in structure):
+                return tf.TensorSpec(
+                    shape=[None] + structure[1:], dtype=model.input_dtype
+                )
+            return [make_tensor_spec(v) for v in structure]
+        else:
+            raise ValueError(
+                f"Unsupported type {type(structure)} for {structure}"
+            )
+
+    return [make_tensor_spec(value) for value in shapes_dict.values()]
+
+
+@keras_export("keras.layers.TFSMLayer")
+class TFSMLayer(Layer):
     """Reload a Keras model/layer that was saved via SavedModel / ExportArchive.
 
     Arguments:
@@ -428,7 +690,7 @@ class ReloadedLayer(base_layer.Layer):
 
     ```python
     model.export("path/to/artifact")
-    reloaded_layer = ReloadedLayer("path/to/artifact")
+    reloaded_layer = TFSMLayer("path/to/artifact")
     outputs = reloaded_layer(inputs)
     ```
 
@@ -443,13 +705,13 @@ class ReloadedLayer(base_layer.Layer):
     * Only call endpoints with a single `inputs` tensor argument
     (which may optionally be a dict/tuple/list of tensors) are supported.
     For endpoints with multiple separate input tensor arguments, consider
-    subclassing `ReloadedLayer` and implementing a `call()` method with a
+    subclassing `TFSMLayer` and implementing a `call()` method with a
     custom signature.
     * If you need training-time behavior to differ from inference-time behavior
     (i.e. if you need the reloaded object to support a `training=True` argument
     in `__call__()`), make sure that the training-time call function is
     saved as a standalone endpoint in the artifact, and provide its name
-    to the `ReloadedLayer` via the `call_training_endpoint` argument.
+    to the `TFSMLayer` via the `call_training_endpoint` argument.
     """
 
     def __init__(
@@ -461,6 +723,12 @@ class ReloadedLayer(base_layer.Layer):
         name=None,
         dtype=None,
     ):
+        if backend.backend() != "tensorflow":
+            raise NotImplementedError(
+                "The TFSMLayer is only currently supported with the "
+                "TensorFlow backend."
+            )
+
         # Initialize an empty layer, then add_weight() etc. as needed.
         super().__init__(trainable=trainable, name=name, dtype=dtype)
 
@@ -479,9 +747,13 @@ class ReloadedLayer(base_layer.Layer):
             self.call_endpoint_fn = self._reloaded_obj.signatures[call_endpoint]
         else:
             raise ValueError(
-                f"The endpoint '{call_endpoint}' is neither an "
-                "attribute of the reloaded SavedModel, nor an entry "
-                "in the `signatures` field of the reloaded SavedModel. "
+                f"The endpoint '{call_endpoint}' "
+                "is neither an attribute of the reloaded SavedModel, "
+                "nor an entry in the `signatures` field of "
+                "the reloaded SavedModel. Select another endpoint via "
+                "the `call_endpoint` argument. Available endpoints for "
+                "this SavedModel: "
+                f"{list(self._reloaded_obj.signatures.keys())}"
             )
 
         # Resolving the training function.
@@ -496,10 +768,12 @@ class ReloadedLayer(base_layer.Layer):
                 ]
             else:
                 raise ValueError(
-                    f"The endpoint '{call_training_endpoint}' is "
-                    "neither an attribute of the reloaded SavedModel, "
+                    f"The endpoint '{call_training_endpoint}' "
+                    "is neither an attribute of the reloaded SavedModel, "
                     "nor an entry in the `signatures` field of "
-                    "the reloaded SavedModel. "
+                    "the reloaded SavedModel. Available endpoints for "
+                    "this SavedModel: "
+                    f"{list(self._reloaded_obj.signatures.keys())}"
                 )
 
         # Add trainable and non-trainable weights from the call_endpoint_fn.
@@ -508,20 +782,14 @@ class ReloadedLayer(base_layer.Layer):
             all_fns.append(self.call_training_endpoint_fn)
         tvs, ntvs = _list_variables_used_by_fns(all_fns)
         for v in tvs:
-            self._add_existing_weight(v, trainable=True)
+            self._add_existing_weight(v)
         for v in ntvs:
-            self._add_existing_weight(v, trainable=False)
+            self._add_existing_weight(v)
         self.built = True
 
-    def _add_existing_weight(self, weight, trainable):
-        """Calls add_weight() to register but not create an existing weight."""
-        self.add_weight(
-            name=weight.name,
-            shape=weight.shape,
-            dtype=weight.dtype,
-            trainable=trainable,
-            getter=lambda *_, **__: weight,
-        )
+    def _add_existing_weight(self, weight):
+        """Tracks an existing weight."""
+        self._track_variable(weight)
 
     def call(self, inputs, training=False, **kwargs):
         if training:
@@ -541,7 +809,8 @@ class ReloadedLayer(base_layer.Layer):
 
 
 def _make_tensor_spec(x):
-    return tf.TensorSpec(x.shape, dtype=x.dtype, name=x.name)
+    shape = (None,) + x.shape[1:]
+    return tf.TensorSpec(shape, dtype=x.dtype, name=x.name)
 
 
 def _print_signature(fn, name):
@@ -579,4 +848,3 @@ def _list_variables_used_by_fns(fns):
                     non_trainable_variables.append(v)
                     non_trainable_variables_ids.add(id(v))
     return trainable_variables, non_trainable_variables
-
