@@ -1,198 +1,73 @@
-# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-#
-#     http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
-# ==============================================================================
-"""Regression metrics, e.g. MAE/MSE/etc."""
-
 import warnings
 
-import tensorflow.compat.v2 as tf
-
-from keras.src import backend
-from keras.src.dtensor import utils as dtensor_utils
-from keras.src.losses import logcosh
-from keras.src.losses import mean_absolute_error
-from keras.src.losses import mean_absolute_percentage_error
-from keras.src.losses import mean_squared_error
-from keras.src.losses import mean_squared_logarithmic_error
-from keras.src.metrics import base_metric
-from keras.src.utils import losses_utils
-from keras.src.utils import metrics_utils
-from keras.src.utils.tf_utils import is_tensor_or_variable
-
-# isort: off
-from tensorflow.python.util.tf_export import keras_export
+from keras.src import initializers
+from keras.src import ops
+from keras.src.api_export import keras_export
+from keras.src.losses.loss import squeeze_or_expand_to_same_rank
+from keras.src.losses.losses import log_cosh
+from keras.src.losses.losses import mean_absolute_error
+from keras.src.losses.losses import mean_absolute_percentage_error
+from keras.src.losses.losses import mean_squared_error
+from keras.src.losses.losses import mean_squared_logarithmic_error
+from keras.src.metrics import reduction_metrics
+from keras.src.utils.numerical_utils import normalize
 
 
-@keras_export("keras.metrics.MeanRelativeError")
-class MeanRelativeError(base_metric.Mean):
-    """Computes the mean relative error by normalizing with the given values.
+@keras_export("keras.metrics.MeanSquaredError")
+class MeanSquaredError(reduction_metrics.MeanMetricWrapper):
+    """Computes the mean squared error between `y_true` and `y_pred`.
 
-    This metric creates two local variables, `total` and `count` that are used
-    to compute the mean relative error. This is weighted by `sample_weight`, and
-    it is ultimately returned as `mean_relative_error`: an idempotent operation
-    that simply divides `total` by `count`.
-
-    If `sample_weight` is `None`, weights default to 1.
-    Use `sample_weight` of 0 to mask values.
-
-    Args:
-      normalizer: The normalizer values with same shape as predictions.
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
-
-    Standalone usage:
-
-    >>> m = tf.keras.metrics.MeanRelativeError(normalizer=[1, 3, 2, 3])
-    >>> m.update_state([1, 3, 2, 3], [2, 4, 6, 8])
-
-    >>> # metric = mean(|y_pred - y_true| / normalizer)
-    >>> #        = mean([1, 1, 4, 5] / [1, 3, 2, 3]) = mean([1, 1/3, 2, 5/3])
-    >>> #        = 5/4 = 1.25
-    >>> m.result().numpy()
-    1.25
-
-    Usage with `compile()` API:
+    Formula:
 
     ```python
-    model.compile(
-      optimizer='sgd',
-      loss='mse',
-      metrics=[tf.keras.metrics.MeanRelativeError(normalizer=[1, 3])])
+    loss = mean(square(y_true - y_pred))
     ```
+
+    Args:
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
+
+    Example:
+
+    >>> m = keras.metrics.MeanSquaredError()
+    >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
+    >>> m.result()
+    0.25
     """
 
-    @dtensor_utils.inject_mesh
-    def __init__(self, normalizer, name=None, dtype=None):
-        super().__init__(name=name, dtype=dtype)
-        normalizer = tf.cast(normalizer, self._dtype)
-        self.normalizer = normalizer
-
-    def update_state(self, y_true, y_pred, sample_weight=None):
-        """Accumulates metric statistics.
-
-        Args:
-          y_true: The ground truth values.
-          y_pred: The predicted values.
-          sample_weight: Optional weighting of each example. Can
-            be a `Tensor` whose rank is either 0, or the same rank as `y_true`,
-            and must be broadcastable to `y_true`. Defaults to `1`.
-
-        Returns:
-          Update op.
-        """
-        y_true = tf.cast(y_true, self._dtype)
-        y_pred = tf.cast(y_pred, self._dtype)
-        [
-            y_pred,
-            y_true,
-        ], sample_weight = metrics_utils.ragged_assert_compatible_and_get_flat_values(  # noqa: E501
-            [y_pred, y_true], sample_weight
-        )
-        y_pred, y_true = losses_utils.squeeze_or_expand_dimensions(
-            y_pred, y_true
-        )
-
-        y_pred, self.normalizer = losses_utils.remove_squeezable_dimensions(
-            y_pred, self.normalizer
-        )
-        y_pred.shape.assert_is_compatible_with(y_true.shape)
-        relative_errors = tf.math.divide_no_nan(
-            tf.abs(y_true - y_pred), self.normalizer
-        )
-
-        return super().update_state(
-            relative_errors, sample_weight=sample_weight
-        )
+    def __init__(self, name="mean_squared_error", dtype=None):
+        super().__init__(fn=mean_squared_error, name=name, dtype=dtype)
+        # Metric should be minimized during optimization.
+        self._direction = "down"
 
     def get_config(self):
-        n = self.normalizer
-        config = {
-            "normalizer": backend.eval(n) if is_tensor_or_variable(n) else n
-        }
-        base_config = super().get_config()
-        return dict(list(base_config.items()) + list(config.items()))
-
-
-@keras_export("keras.metrics.CosineSimilarity")
-class CosineSimilarity(base_metric.MeanMetricWrapper):
-    """Computes the cosine similarity between the labels and predictions.
-
-    `cosine similarity = (a . b) / ||a|| ||b||`
-
-    See: [Cosine Similarity](https://en.wikipedia.org/wiki/Cosine_similarity).
-
-    This metric keeps the average cosine similarity between `predictions` and
-    `labels` over a stream of data.
-
-    Args:
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
-      axis: (Optional) The dimension along which the cosine
-        similarity is computed. Defaults to `-1`.
-
-    Standalone usage:
-
-    >>> # l2_norm(y_true) = [[0., 1.], [1./1.414, 1./1.414]]
-    >>> # l2_norm(y_pred) = [[1., 0.], [1./1.414, 1./1.414]]
-    >>> # l2_norm(y_true) . l2_norm(y_pred) = [[0., 0.], [0.5, 0.5]]
-    >>> # result = mean(sum(l2_norm(y_true) . l2_norm(y_pred), axis=1))
-    >>> #        = ((0. + 0.) +  (0.5 + 0.5)) / 2
-    >>> m = tf.keras.metrics.CosineSimilarity(axis=1)
-    >>> m.update_state([[0., 1.], [1., 1.]], [[1., 0.], [1., 1.]])
-    >>> m.result().numpy()
-    0.49999997
-
-    >>> m.reset_state()
-    >>> m.update_state([[0., 1.], [1., 1.]], [[1., 0.], [1., 1.]],
-    ...                sample_weight=[0.3, 0.7])
-    >>> m.result().numpy()
-    0.6999999
-
-    Usage with `compile()` API:
-
-    ```python
-    model.compile(
-        optimizer='sgd',
-        loss='mse',
-        metrics=[tf.keras.metrics.CosineSimilarity(axis=1)])
-    ```
-    """
-
-    @dtensor_utils.inject_mesh
-    def __init__(self, name="cosine_similarity", dtype=None, axis=-1):
-        super().__init__(cosine_similarity, name, dtype=dtype, axis=axis)
+        return {"name": self.name, "dtype": self.dtype}
 
 
 @keras_export("keras.metrics.MeanAbsoluteError")
-class MeanAbsoluteError(base_metric.MeanMetricWrapper):
+class MeanAbsoluteError(reduction_metrics.MeanMetricWrapper):
     """Computes the mean absolute error between the labels and predictions.
 
+    Formula:
+
+    ```python
+    loss = mean(abs(y_true - y_pred))
+    ```
+
     Args:
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
 
-    Standalone usage:
+    Examples:
 
-    >>> m = tf.keras.metrics.MeanAbsoluteError()
+    >>> m = keras.metrics.MeanAbsoluteError()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
-    >>> m.result().numpy()
+    >>> m.result()
     0.25
-
     >>> m.reset_state()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]],
     ...                sample_weight=[1, 0])
-    >>> m.result().numpy()
+    >>> m.result()
     0.5
 
     Usage with `compile()` API:
@@ -201,35 +76,45 @@ class MeanAbsoluteError(base_metric.MeanMetricWrapper):
     model.compile(
         optimizer='sgd',
         loss='mse',
-        metrics=[tf.keras.metrics.MeanAbsoluteError()])
+        metrics=[keras.metrics.MeanAbsoluteError()])
     ```
     """
 
-    @dtensor_utils.inject_mesh
     def __init__(self, name="mean_absolute_error", dtype=None):
         super().__init__(mean_absolute_error, name, dtype=dtype)
+        # Metric should be minimized during optimization.
+        self._direction = "down"
+
+    def get_config(self):
+        return {"name": self.name, "dtype": self.dtype}
 
 
 @keras_export("keras.metrics.MeanAbsolutePercentageError")
-class MeanAbsolutePercentageError(base_metric.MeanMetricWrapper):
-    """Computes the mean absolute percentage error between `y_true` and
-    `y_pred`.
+class MeanAbsolutePercentageError(reduction_metrics.MeanMetricWrapper):
+    """Computes mean absolute percentage error between `y_true` and `y_pred`.
+
+    Formula:
+
+    ```python
+    loss = 100 * mean(abs((y_true - y_pred) / y_true))
+    ```
 
     Args:
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
 
-    Standalone usage:
+    Example:
 
-    >>> m = tf.keras.metrics.MeanAbsolutePercentageError()
+    Example:
+
+    >>> m = keras.metrics.MeanAbsolutePercentageError()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
-    >>> m.result().numpy()
+    >>> m.result()
     250000000.0
-
     >>> m.reset_state()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]],
     ...                sample_weight=[1, 0])
-    >>> m.result().numpy()
+    >>> m.result()
     500000000.0
 
     Usage with `compile()` API:
@@ -238,71 +123,45 @@ class MeanAbsolutePercentageError(base_metric.MeanMetricWrapper):
     model.compile(
         optimizer='sgd',
         loss='mse',
-        metrics=[tf.keras.metrics.MeanAbsolutePercentageError()])
+        metrics=[keras.metrics.MeanAbsolutePercentageError()])
     ```
     """
 
-    @dtensor_utils.inject_mesh
     def __init__(self, name="mean_absolute_percentage_error", dtype=None):
         super().__init__(mean_absolute_percentage_error, name, dtype=dtype)
+        # Metric should be minimized during optimization.
+        self._direction = "down"
 
-
-@keras_export("keras.metrics.MeanSquaredError")
-class MeanSquaredError(base_metric.MeanMetricWrapper):
-    """Computes the mean squared error between `y_true` and `y_pred`.
-
-    Args:
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
-
-    Standalone usage:
-
-    >>> m = tf.keras.metrics.MeanSquaredError()
-    >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
-    >>> m.result().numpy()
-    0.25
-
-    >>> m.reset_state()
-    >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]],
-    ...                sample_weight=[1, 0])
-    >>> m.result().numpy()
-    0.5
-
-    Usage with `compile()` API:
-
-    ```python
-    model.compile(
-        optimizer='sgd',
-        loss='mse',
-        metrics=[tf.keras.metrics.MeanSquaredError()])
-    ```
-    """
-
-    @dtensor_utils.inject_mesh
-    def __init__(self, name="mean_squared_error", dtype=None):
-        super().__init__(mean_squared_error, name, dtype=dtype)
+    def get_config(self):
+        return {"name": self.name, "dtype": self.dtype}
 
 
 @keras_export("keras.metrics.MeanSquaredLogarithmicError")
-class MeanSquaredLogarithmicError(base_metric.MeanMetricWrapper):
-    """Computes the mean squared logarithmic error between `y_true` and
-    `y_pred`.
+class MeanSquaredLogarithmicError(reduction_metrics.MeanMetricWrapper):
+    """Computes mean squared logarithmic error between `y_true` and `y_pred`.
+
+    Formula:
+
+    ```python
+    loss = mean(square(log(y_true + 1) - log(y_pred + 1)))
+    ```
 
     Args:
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
 
-    Standalone usage:
+    Example:
 
-    >>> m = tf.keras.metrics.MeanSquaredLogarithmicError()
+    Example:
+
+    >>> m = keras.metrics.MeanSquaredLogarithmicError()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
-    >>> m.result().numpy()
+    >>> m.result()
     0.12011322
-
     >>> m.reset_state()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]],
     ...                sample_weight=[1, 0])
-    >>> m.result().numpy()
+    >>> m.result()
     0.24022643
 
     Usage with `compile()` API:
@@ -311,30 +170,46 @@ class MeanSquaredLogarithmicError(base_metric.MeanMetricWrapper):
     model.compile(
         optimizer='sgd',
         loss='mse',
-        metrics=[tf.keras.metrics.MeanSquaredLogarithmicError()])
+        metrics=[keras.metrics.MeanSquaredLogarithmicError()])
     ```
     """
 
-    @dtensor_utils.inject_mesh
     def __init__(self, name="mean_squared_logarithmic_error", dtype=None):
         super().__init__(mean_squared_logarithmic_error, name, dtype=dtype)
+        # Metric should be minimized during optimization.
+        self._direction = "down"
+
+    def get_config(self):
+        return {"name": self.name, "dtype": self.dtype}
 
 
 @keras_export("keras.metrics.RootMeanSquaredError")
-class RootMeanSquaredError(base_metric.Mean):
+class RootMeanSquaredError(reduction_metrics.Mean):
     """Computes root mean squared error metric between `y_true` and `y_pred`.
 
-    Standalone usage:
+    Formula:
 
-    >>> m = tf.keras.metrics.RootMeanSquaredError()
+    ```python
+    loss = sqrt(mean((y_pred - y_true) ** 2))
+    ```
+
+    Args:
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
+
+    Example:
+
+    Example:
+
+    >>> m = keras.metrics.RootMeanSquaredError()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
-    >>> m.result().numpy()
+    >>> m.result()
     0.5
 
     >>> m.reset_state()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]],
     ...                sample_weight=[1, 0])
-    >>> m.result().numpy()
+    >>> m.result()
     0.70710677
 
     Usage with `compile()` API:
@@ -343,61 +218,123 @@ class RootMeanSquaredError(base_metric.Mean):
     model.compile(
         optimizer='sgd',
         loss='mse',
-        metrics=[tf.keras.metrics.RootMeanSquaredError()])
+        metrics=[keras.metrics.RootMeanSquaredError()])
     ```
     """
 
-    @dtensor_utils.inject_mesh
     def __init__(self, name="root_mean_squared_error", dtype=None):
         super().__init__(name, dtype=dtype)
+        # Metric should be minimized during optimization.
+        self._direction = "down"
 
     def update_state(self, y_true, y_pred, sample_weight=None):
         """Accumulates root mean squared error statistics.
 
         Args:
-          y_true: The ground truth values.
-          y_pred: The predicted values.
-          sample_weight: Optional weighting of each example. Can
-            be a `Tensor` whose rank is either 0, or the same rank as `y_true`,
-            and must be broadcastable to `y_true`. Defaults to `1`.
+            y_true: The ground truth values.
+            y_pred: The predicted values.
+            sample_weight: Optional weighting of each example. Can
+                be a `Tensor` whose rank is either 0, or the same rank as
+                `y_true`, and must be broadcastable to `y_true`.
+                Defaults to `1`.
 
         Returns:
-          Update op.
+            Update op.
         """
-        y_true = tf.cast(y_true, self._dtype)
-        y_pred = tf.cast(y_pred, self._dtype)
-        y_pred, y_true = losses_utils.squeeze_or_expand_dimensions(
-            y_pred, y_true
-        )
-        error_sq = tf.math.squared_difference(y_pred, y_true)
+        y_true = ops.convert_to_tensor(y_true, self._dtype)
+        y_pred = ops.convert_to_tensor(y_pred, self._dtype)
+        y_true, y_pred = squeeze_or_expand_to_same_rank(y_true, y_pred)
+        error_sq = ops.square(y_pred - y_true)
         return super().update_state(error_sq, sample_weight=sample_weight)
 
     def result(self):
-        return tf.sqrt(tf.math.divide_no_nan(self.total, self.count))
+        return ops.sqrt(super().result())
+
+
+@keras_export("keras.metrics.CosineSimilarity")
+class CosineSimilarity(reduction_metrics.MeanMetricWrapper):
+    """Computes the cosine similarity between the labels and predictions.
+
+    Formula:
+
+    ```python
+    loss = sum(l2_norm(y_true) * l2_norm(y_pred))
+    ```
+    See: [Cosine Similarity](https://en.wikipedia.org/wiki/Cosine_similarity).
+    This metric keeps the average cosine similarity between `predictions` and
+    `labels` over a stream of data.
+
+    Args:
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
+        axis: (Optional) Defaults to `-1`. The dimension along which the cosine
+            similarity is computed.
+
+    Example:
+
+    Example:
+
+    >>> # l2_norm(y_true) = [[0., 1.], [1./1.414, 1./1.414]]
+    >>> # l2_norm(y_pred) = [[1., 0.], [1./1.414, 1./1.414]]
+    >>> # l2_norm(y_true) . l2_norm(y_pred) = [[0., 0.], [0.5, 0.5]]
+    >>> # result = mean(sum(l2_norm(y_true) . l2_norm(y_pred), axis=1))
+    >>> #        = ((0. + 0.) +  (0.5 + 0.5)) / 2
+    >>> m = keras.metrics.CosineSimilarity(axis=1)
+    >>> m.update_state([[0., 1.], [1., 1.]], [[1., 0.], [1., 1.]])
+    >>> m.result()
+    0.49999997
+    >>> m.reset_state()
+    >>> m.update_state([[0., 1.], [1., 1.]], [[1., 0.], [1., 1.]],
+    ...                sample_weight=[0.3, 0.7])
+    >>> m.result()
+    0.6999999
+
+    Usage with `compile()` API:
+
+    ```python
+    model.compile(
+        optimizer='sgd',
+        loss='mse',
+        metrics=[keras.metrics.CosineSimilarity(axis=1)])
+    ```
+    """
+
+    def __init__(self, name="cosine_similarity", dtype=None, axis=-1):
+        super().__init__(cosine_similarity, name, dtype=dtype, axis=axis)
+        # Metric should be maximized during optimization.
+        self._direction = "up"
+
+    def get_config(self):
+        return {"name": self.name, "dtype": self.dtype}
 
 
 @keras_export("keras.metrics.LogCoshError")
-class LogCoshError(base_metric.MeanMetricWrapper):
+class LogCoshError(reduction_metrics.MeanMetricWrapper):
     """Computes the logarithm of the hyperbolic cosine of the prediction error.
 
-    `logcosh = log((exp(x) + exp(-x))/2)`, where x is the error (y_pred -
-    y_true)
+    Formula:
+
+    ```python
+    error = y_pred - y_true
+    logcosh = mean(log((exp(error) + exp(-error))/2), axis=-1)
+    ```
 
     Args:
-      name: (Optional) string name of the metric instance.
-      dtype: (Optional) data type of the metric result.
+        name: (Optional) string name of the metric instance.
+        dtype: (Optional) data type of the metric result.
 
-    Standalone usage:
+    Example:
 
-    >>> m = tf.keras.metrics.LogCoshError()
+    Example:
+
+    >>> m = keras.metrics.LogCoshError()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]])
-    >>> m.result().numpy()
+    >>> m.result()
     0.10844523
-
     >>> m.reset_state()
     >>> m.update_state([[0, 1], [0, 0]], [[1, 1], [0, 0]],
     ...                sample_weight=[1, 0])
-    >>> m.result().numpy()
+    >>> m.result()
     0.21689045
 
     Usage with `compile()` API:
@@ -405,23 +342,35 @@ class LogCoshError(base_metric.MeanMetricWrapper):
     ```python
     model.compile(optimizer='sgd',
                   loss='mse',
-                  metrics=[tf.keras.metrics.LogCoshError()])
+                  metrics=[keras.metrics.LogCoshError()])
     ```
     """
 
-    @dtensor_utils.inject_mesh
     def __init__(self, name="logcosh", dtype=None):
-        super().__init__(logcosh, name, dtype=dtype)
+        super().__init__(log_cosh, name, dtype=dtype)
+        # Metric should be minimized during optimization.
+        self._direction = "down"
+
+    def get_config(self):
+        return {"name": self.name, "dtype": self.dtype}
 
 
 # Adapted from TF-Addons implementation (RSquare class).
 @keras_export("keras.metrics.R2Score")
-class R2Score(base_metric.Metric):
+class R2Score(reduction_metrics.Metric):
     """Computes R2 score.
 
+    Formula:
+
+    ```python
+    sum_squares_residuals = sum((y_true - y_pred) ** 2)
+    sum_squares = sum((y_true - mean(y_true)) ** 2)
+    R2 = 1 - sum_squares_residuals / sum_squares
+    ```
+
     This is also called the
-    [coefficient of
-    determination](https://en.wikipedia.org/wiki/Coefficient_of_determination).
+    [coefficient of determination](
+    https://en.wikipedia.org/wiki/Coefficient_of_determination).
 
     It indicates how close the fitted regression line
     is to ground-truth data.
@@ -452,14 +401,13 @@ class R2Score(base_metric.Metric):
 
     >>> y_true = np.array([[1], [4], [3]], dtype=np.float32)
     >>> y_pred = np.array([[2], [4], [4]], dtype=np.float32)
-    >>> metric = tf.keras.metrics.R2Score()
+    >>> metric = keras.metrics.R2Score()
     >>> metric.update_state(y_true, y_pred)
     >>> result = metric.result()
-    >>> result.numpy()
+    >>> result
     0.57142854
     """
 
-    @dtensor_utils.inject_mesh
     def __init__(
         self,
         class_aggregation="uniform_average",
@@ -468,6 +416,8 @@ class R2Score(base_metric.Metric):
         dtype=None,
     ):
         super().__init__(name=name, dtype=dtype)
+        # Metric should be maximized during optimization.
+        self._direction = "up"
 
         valid_class_aggregation_values = (
             None,
@@ -488,10 +438,14 @@ class R2Score(base_metric.Metric):
             )
         self.class_aggregation = class_aggregation
         self.num_regressors = num_regressors
-        self.num_samples = self.add_weight(name="num_samples", dtype="int32")
-        self.built = False
+        self.num_samples = self.add_variable(
+            shape=(),
+            initializer=initializers.Zeros(),
+            name="num_samples",
+        )
+        self._built = False
 
-    def build(self, y_true_shape, y_pred_shape):
+    def _build(self, y_true_shape, y_pred_shape):
         if len(y_pred_shape) != 2 or len(y_true_shape) != 2:
             raise ValueError(
                 "R2Score expects 2D inputs with shape "
@@ -508,67 +462,85 @@ class R2Score(base_metric.Metric):
                 f"y_true.shape={y_true_shape}."
             )
         num_classes = y_pred_shape[-1]
-        self.squared_sum = self.add_weight(
+        self.squared_sum = self.add_variable(
             name="squared_sum",
             shape=[num_classes],
-            initializer="zeros",
+            initializer=initializers.Zeros(),
         )
-        self.sum = self.add_weight(
+        self.sum = self.add_variable(
             name="sum",
             shape=[num_classes],
-            initializer="zeros",
+            initializer=initializers.Zeros(),
         )
-        self.total_mse = self.add_weight(
+        self.total_mse = self.add_variable(
             name="residual",
             shape=[num_classes],
-            initializer="zeros",
+            initializer=initializers.Zeros(),
         )
-        self.count = self.add_weight(
+        self.count = self.add_variable(
             name="count",
             shape=[num_classes],
-            initializer="zeros",
+            initializer=initializers.Zeros(),
         )
-        self.built = True
+        self._built = True
 
     def update_state(self, y_true, y_pred, sample_weight=None):
-        y_true = tf.convert_to_tensor(y_true, dtype=self.dtype)
-        y_pred = tf.convert_to_tensor(y_pred, dtype=self.dtype)
-        if not self.built:
-            self.build(y_true.shape, y_pred.shape)
+        """Accumulates root mean squared error statistics.
+
+        Args:
+            y_true: The ground truth values.
+            y_pred: The predicted values.
+            sample_weight: Optional weighting of each example. Can
+                be a `Tensor` whose rank is either 0, or the same rank as
+                `y_true`, and must be broadcastable to `y_true`.
+                Defaults to `1`.
+
+        Returns:
+            Update op.
+        """
+        y_true = ops.convert_to_tensor(y_true, dtype=self._dtype)
+        y_pred = ops.convert_to_tensor(y_pred, dtype=self._dtype)
+        y_true, y_pred = squeeze_or_expand_to_same_rank(y_true, y_pred)
+        if not self._built:
+            self._build(y_true.shape, y_pred.shape)
 
         if sample_weight is None:
             sample_weight = 1
 
-        sample_weight = tf.convert_to_tensor(sample_weight, dtype=self.dtype)
-        if sample_weight.shape.rank == 1:
-            # Make sure there's a features dimension
-            sample_weight = tf.expand_dims(sample_weight, axis=1)
-        sample_weight = tf.__internal__.ops.broadcast_weights(
-            weights=sample_weight, values=y_true
-        )
+        sample_weight = ops.convert_to_tensor(sample_weight, dtype=self.dtype)
 
-        weighted_y_true = y_true * sample_weight
-        self.sum.assign_add(tf.reduce_sum(weighted_y_true, axis=0))
-        self.squared_sum.assign_add(
-            tf.reduce_sum(y_true * weighted_y_true, axis=0)
+        if len(sample_weight.shape) == 1:
+            # Make sure there's a features dimension
+            sample_weight = ops.expand_dims(sample_weight, axis=1)
+
+        sample_weight = ops.broadcast_to(sample_weight, ops.shape(y_true))
+
+        weighted_y_true = y_true * ops.cast(sample_weight, y_true.dtype)
+        self.sum.assign(self.sum + ops.sum(weighted_y_true, axis=0))
+        self.squared_sum.assign(
+            self.squared_sum + ops.sum(y_true * weighted_y_true, axis=0)
         )
-        self.total_mse.assign_add(
-            tf.reduce_sum((y_true - y_pred) ** 2 * sample_weight, axis=0)
+        self.total_mse.assign(
+            self.total_mse
+            + ops.sum(
+                (y_true - y_pred) ** 2 * ops.cast(sample_weight, y_true.dtype),
+                axis=0,
+            )
         )
-        self.count.assign_add(tf.reduce_sum(sample_weight, axis=0))
-        self.num_samples.assign_add(tf.size(y_true))
+        self.count.assign(self.count + ops.sum(sample_weight, axis=0))
+        self.num_samples.assign(self.num_samples + ops.size(y_true))
 
     def result(self):
         mean = self.sum / self.count
         total = self.squared_sum - self.sum * mean
         raw_scores = 1 - (self.total_mse / total)
-        raw_scores = tf.where(tf.math.is_inf(raw_scores), 0.0, raw_scores)
+        raw_scores = ops.where(ops.isinf(raw_scores), 0.0, raw_scores)
 
         if self.class_aggregation == "uniform_average":
-            r2_score = tf.reduce_mean(raw_scores)
+            r2_score = ops.mean(raw_scores)
         elif self.class_aggregation == "variance_weighted_average":
-            weighted_sum = tf.reduce_sum(total * raw_scores)
-            sum_of_weights = tf.reduce_sum(total)
+            weighted_sum = ops.sum(total * raw_scores)
+            sum_of_weights = ops.sum(total)
             r2_score = weighted_sum / sum_of_weights
         else:
             r2_score = raw_scores
@@ -587,21 +559,23 @@ class R2Score(base_metric.Metric):
                     stacklevel=2,
                 )
             else:
-                n = tf.cast(self.num_samples, dtype=tf.float32)
-                p = tf.cast(self.num_regressors, dtype=tf.float32)
-                num = tf.multiply(
-                    tf.subtract(1.0, r2_score), tf.subtract(n, 1.0)
+                n = ops.convert_to_tensor(self.num_samples, dtype="float32")
+                p = ops.convert_to_tensor(self.num_regressors, dtype="float32")
+                num = ops.multiply(
+                    ops.subtract(1.0, r2_score), ops.subtract(n, 1.0)
                 )
-                den = tf.subtract(tf.subtract(n, p), 1.0)
-                r2_score = tf.subtract(1.0, tf.divide(num, den))
+                den = ops.subtract(ops.subtract(n, p), 1.0)
+                r2_score = ops.subtract(1.0, ops.divide(num, den))
         return r2_score
 
     def reset_state(self):
         for v in self.variables:
-            v.assign(tf.zeros(v.shape, dtype=v.dtype))
+            v.assign(ops.zeros(v.shape, dtype=v.dtype))
 
     def get_config(self):
         config = {
+            "name": self.name,
+            "dtype": self.dtype,
             "class_aggregation": self.class_aggregation,
             "num_regressors": self.num_regressors,
         }
@@ -612,16 +586,30 @@ class R2Score(base_metric.Metric):
 def cosine_similarity(y_true, y_pred, axis=-1):
     """Computes the cosine similarity between labels and predictions.
 
+    Formula:
+
+    ```python
+    loss = sum(l2_norm(y_true) * l2_norm(y_pred))
+    ```
+
     Args:
-      y_true: The ground truth values.
-      y_pred: The prediction values.
-      axis: (Optional) -1 is the dimension along which the cosine
-        similarity is computed. Defaults to `-1`.
+        y_true: Tensor of true targets.
+        y_pred: Tensor of predicted targets.
+        axis: Axis along which to determine similarity. Defaults to `-1`.
 
     Returns:
-      Cosine similarity value.
-    """
-    y_true = tf.linalg.l2_normalize(y_true, axis=axis)
-    y_pred = tf.linalg.l2_normalize(y_pred, axis=axis)
-    return tf.reduce_sum(y_true * y_pred, axis=axis)
+        Cosine similarity tensor.
 
+    Example:
+
+    >>> y_true = [[0., 1.], [1., 1.], [1., 1.]]
+    >>> y_pred = [[1., 0.], [1., 1.], [-1., -1.]]
+    >>> loss = keras.losses.cosine_similarity(y_true, y_pred, axis=-1)
+    [0., 0.99999994, -0.99999994]
+    """
+    y_pred = ops.convert_to_tensor(y_pred)
+    y_true = ops.convert_to_tensor(y_true, dtype=y_pred.dtype)
+    y_true, y_pred = squeeze_or_expand_to_same_rank(y_true, y_pred)
+    y_pred = normalize(y_pred, axis=axis)
+    y_true = normalize(y_true, axis=axis)
+    return ops.sum(y_true * y_pred, axis=axis)
